"""This module holds classes and functions related to heaps and priority queues.

This module contains `PriorityQueue`, an implementation of a priority queue, and `PriorityQueueUnderflowError`, an error
raised by `PriorityQueue`. This module also contains the function `build_min_heap` which modifies the elements of an
`dalpy.arrays.Array` so that they construct a min heap.

Examples:
    Initializing, adding, and removing elements from a `PriorityQueue`:

        q = PriorityQueue()
        q.insert('a', 2)
        q.insert('b', 1)
        q.extract_min()

    The following code will raise a `PriorityQueueUnderflowError` because the second extract min is done on an empty
    `PriorityQueue`:

        q.extract_min()
        q.extract_min()
"""

from dalpy.arrays import Array
from heapq import heapify


class PriorityQueueUnderflowError(Exception):
    """This class is used by `PriorityQueue` to raise errors for operations done on an empty `PriorityQueue`."""

    def __init__(self, operation):
        """Initializes a `PriorityQueueUnderflowError` that will be raised associated with a `PriorityQueue` operation.

        Args:
            operation: a string specifying the operation to raise an error on
        """
        super().__init__(f'Cannot perform {operation} on an empty priority queue.')


class PriorityQueue:
    """This class represents a minimum priority queue.

    One may assume that this `PriorityQueue` has no maximum capacity.

    Examples:
        To initialize a `PriorityQueue`:

            q = PriorityQueue()

        To add elements to `q`:

            q.insert('a', 2)
            q.insert('b', 1)

        To remove and return the minimum priority element of `q` (in this case `x = 'b'`):

            x = q.extract_min()

        To see the minimum priority element of `q` (in this case `y = 'a'`):

            y = q.front()

        To decrease the priority of an element in `q`:

            q.decrease_key('a', 0)
    """

    def __init__(self):
        """Initializes an empty `PriorityQueue` in `O(1)` time."""
        self.__buf = list()

    def insert(self, element, priority):
        """Inserts an element into the `PriorityQueue` with an associated priority.

        One may assume that this operation runs in `O(log(n))` time where `n` is the size of this `Queue`.

        Args:
            element: An element to add to this `PriorityQueue`. This can be of any type.
            priority: The integer priority `element` should have in this `PriorityQueue`.
        """
        i = 0
        while i < len(self.__buf) and self.__buf[i][0] <= priority:
            i += 1
        self.__buf.insert(i, (priority, element))

    def extract_min(self):
        """Removes the minimum priority element of this `PriorityQueue`.

        One may assume that this operation runs in `O(log(n))` time where `n` is the size of this `PriorityQueue`.

        Returns:
            The element with the minimum priority in this `PriorityQueue`.

        Raises:
             PriorityQueueUnderflowError: If this `PriorityQueue` is empty.
        """
        if len(self.__buf) == 0:
            raise PriorityQueueUnderflowError('extract_min()')
        return self.__buf.pop(0)[1]

    def minimum(self):
        """Gets the minimum priority element of this `PriorityQueue`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `PriorityQueue`.

        Returns:
            The element with the minimum priority in this `PriorityQueue`.

        Raises:
             PriorityQueueUnderflowError: If this `PriorityQueue` is empty.
        """
        if len(self.__buf) == 0:
            raise PriorityQueueUnderflowError('minimum()')
        return self.__buf[0][1]

    def decrease_key(self, element, new_priority):
        """Decreases the priority of an element in this `PriorityQueue`.

        One may assume that this operation runs in `O(log(n))` time where `n` is the size of this `PriorityQueue`.

        Args:
            element: The element whose priority is being updated.
            new_priority: The new priority of `element`. It should be `<=` its existing priority.

        Raises:
            ValueError: If `element` is not in this `PriorityQueue` or `new_priority` is greater than the existing
                        priority of `element`.
        """
        idx = -1
        for i, (_, e) in enumerate(self.__buf):
            if e == element:
                idx = i
                break
        if idx == -1:
            raise ValueError(f'{element} does not exist in PriorityQueue')
        priority = self.__buf[idx][0]
        if new_priority > priority:
            raise ValueError(f'new priority {new_priority} must be <= current priority {priority}')
        if new_priority < priority:
            self.__buf.pop(idx)
            self.insert(element, new_priority)

    def size(self):
        """Returns the size of this `PriorityQueue` in `O(1)` time w/r/t the size of this `PriorityQueue`.

        Returns:
            The integer number of elements in this `PriorityQueue`.
        """
        return len(self.__buf)

    def is_empty(self):
        """Returns whether this `PriorityQueue` is empty in `O(1)` time w/r/t the size of this `PriorityQueue`.

        Returns:
            `True` if this `PriorityQueue` is empty, `False` otherwise.
        """
        return len(self.__buf) == 0


def build_min_heap(arr):
    """Modifies an `dalpy.arrays.Array` so that its elements make up a min heap.

    This method does not return a copy of the provided `dalpy.arrays.Array` whose elements make up a heap, it modifies it in place.
    Furthermore, all the elements (starting from index 0) are in a min heap. A user may assume that this method runs in
    `O(n)` time where `n` is the length of the input `dalpy.arrays.Array`.

    Args:
        arr: The input `dalpy.arrays.Array`. Its elements should be comparable with `<`, `>=`, etc.

    Raises:
        TypeError: If `arr` is not an `dalpy.arrays.Array`'.
    """
    if not isinstance(arr, Array):
        raise TypeError('can only build min heap of an Array')
    ls = [arr[i] for i in range(arr.length())]
    heapify(ls)
    for i in range(arr.length()):
        arr[i] = ls[i]



"""DALPy is a Python module for learning data structures and algorithms. It is based off of *Introduction to Algorithms*
by Thomas H. Cormen, Charles E. Leiserson, Ronald L. Rivest, and Clifford Stein. This library was made specifically for
administering and grading assignments related to data structures and algorithms in computer science.

With this module students can receive progress reports on their problem sets in real time as they complete assignments.
Additionally, student submission assessment is done with unit tests, instead of hand-tracing, ensuring that the grades
that students receive accurately reflect their submissions.

The DALPy testing suite offers extremely lightweight and flexible unit testing utilities that can be used on any
kind of assignment, whether to write functions or build classes. Course administration can be easily streamlined by
restricting which library data structures students are allowed to use on any particular assignment.

DALPy began as a project by two Brandeis University undergraduate students to replace hand-written problem sets
written in pseudocode.

## Provided Data Structures
The DALPy library offers a set of fundamental data structures and algorithms, with behavior as specified by H.
Cormen et al.'s *Introduction to Algorithms*. The following structures (separated by module) are supported:

* [arrays](https://dalpy-developers.github.io/DALPy/arrays.html)
    * Array
    * Array2D
* [queues](https://dalpy-developers.github.io/DALPy/queues.html)
    * Queue
* [stacks](https://dalpy-developers.github.io/DALPy/stacks.html)
    * Stack
* [sets](https://dalpy-developers.github.io/DALPy/sets.html)
    * Set
* [linked_lists](https://dalpy-developers.github.io/DALPy/linked_lists.html)
    * SinglyLinkedListNode
    * DoublyLinkedListNode
* [trees](https://dalpy-developers.github.io/DALPy/trees.html)
    * BinaryTreeNode
    * NaryTreeNode
    * depth(NaryTreeNode)
* [hashing](https://dalpy-developers.github.io/DALPy/hashing.html)
    * HashTable
* [heaps](https://dalpy-developers.github.io/DALPy/heaps.html)
    * PriorityQueue
    * build_min_heap(Array)
* [graphs](https://dalpy-developers.github.io/DALPy/graphs.html)
    * Vertex
    * Graph

## Unit Testing
Along with the DALPy data structures come test utilities for writing test cases. The testing framework allows a
course administrator to easily write test cases for either expected function output or general class behavior. Test
cases can then be compiled into a testing suite. The testing suite has the capability to set a test case run-time 
timeout and to record comma-separated test results for administrative use.

Consider the example test case below:

    import unittest
    from dalpy.factory_utils import make_stack
    from dalpy.test_utils import build_and_run_watched_suite, generic_test

    from student_submission import student_function

    # TestCase class for testing student_function
    class StudentFunctionTest(unittest.TestCase):

        # A single test case
        def simple_test_case(self):
            stack = make_stack([1, 2, 3])
            expected = make_stack([1, 1, 2, 2, 3, 3])
            generic_test(stack, expected, student_function, in_place=True)

    # Run the test cases using build_and_run_watched_suite with a timeout of 4 seconds
    if __name__ == '__main__':
        build_and_run_watched_suite([StudentFunctionTest], 4)


## Installation

DALPy is [available on PyPI](https://pypi.org/project/dalpy/), and can be installed with pip.

    pip install dalpy

DALPy has the following dependencies:

    Python >= 3.6

## Issues

We encourage you to report issues using the GitHub tracker. We welcome all kinds of issues, especially those related to
correctness, documentation and feature requests.

## Academic Usage

If you are planning to use DALPy for a university course and have questions, feel free to reach out by email.

## Documentation

The full documentation for DALPy is available [here](https://dalpy-developers.github.io/DALPy/).

## Sample Usage

To view sample assignments using DALPy browse the [DALPy sample problems repository](https://github.com/DALPy-Developers/DALPy-Sample-Problems) on GitHub.

## Notes

This project was formerly known as Cormen-Lib.
"""


"""This module holds classes related to LIFO stacks.

This module contains `Stack`, an implementation of a LIFO stack, and `StackUnderflowError`, an error raised by `Stack`.

Examples:
    Initializing, adding, and removing elements from a `Stack`:

        s = Stack()
        s.push(1)
        s.push(2)
        s.pop()

    The following code will raise a `StackUnderflowError` because the second pop is done on an empty `Stack`:

        s.pop()
        s.pop()
"""


class StackUnderflowError(Exception):
    """This class is used by `Stack` to raise errors for operations done on an empty `Stack`."""

    def __init__(self, operation):
        """Initializes a `StackUnderflowError` that will be raised associated with a particular `Stack` operation.

        Args:
            operation: a string specifying the operation to raise an error on
        """
        super().__init__(f'Cannot perform {operation} on an empty stack.')


class Stack:
    """This class represents a LIFO stack.

    One may assume that this `Stack` has no maximum capacity.

    Examples:
        To initialize a `Stack`:

            s = Stack()

        To add elements to the end of `s`:

            s.push(1)
            s.push(2)

        To remove and return the element at the top of `s` (in this case `x = 2`):

            x = s.pop()

        To see the element at the top of `s` (in this case `y = 1`):

            y = s.top()
    """

    def __init__(self):
        """Initializes an empty `Stack` in `O(1)` time."""
        self.__buf = list()

    def top(self):
        """Gets the element at the top of this `Stack`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `Stack`.

        Returns:
            The element at the top of the `Stack`. That is, the element that was last added to this `Stack` of the
            elements in it.

        Raises:
             StackUnderflowError: If this `Stack` is empty.
        """
        if len(self.__buf) == 0:
            raise StackUnderflowError('top()')
        return self.__buf[-1]

    def pop(self):
        """Removes the element at the top of this `Stack`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `Stack`.

        Returns:
            The element at the top of the `Stack` that was removed. That is, the element that was last added to this
            `Stack` of the elements in it.

        Raises:
             StackUnderflowError: If this `Stack` is empty.
        """
        if len(self.__buf) == 0:
            raise StackUnderflowError('pop()')
        return self.__buf.pop()

    def push(self, value):
        """Adds an element to the top of this `Stack`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `Stack`.

        Args:
            value: Element to add to this `Stack`. It can be of any type.
        """
        self.__buf.append(value)

    def is_empty(self):
        """Returns `True` if this `Stack` is empty, `False` otherwise in `O(1)` time w/r/t the size of this `Stack`."""
        return len(self.__buf) == 0

    def size(self):
        """Returns the integer number of elements in this `Stack` in `O(1)` time w/r/t the size of this `Stack`."""
        return len(self.__buf)



"""This module provides miscellaneous utility functions and constants.

Currently, this module holds utilities only related to mathematical operations, but operations related to other aspects
of Python that should be wrapped into DALPy could be added here (e.g. related to strings).
"""

import sys
import math

INF = sys.maxsize
"""Integer constant representing infinity."""

NEG_INF = -sys.maxsize
"""Integer constant representing negative infinity."""


def ceil(x):
    """Returns x rounded up to the nearest integer."""
    return math.ceil(x)


def floor(x):
    """Returns x truncated."""
    return math.floor(x)



"""Module that holds classes and functions related to trees.

This module contains the `BinaryTreeNode` and `NaryTreeNode` classes. `BinaryTreeNode` can be used to recursively define
a binary tree. `NaryTreeNode` can be used to recursively define a doubly linked list. Problems or algorithms that
operate on trees should take an object of either of these classes. This module also contains the `depth` function which
computes the depth of an `NaryTreeNode` in a tree.

Examples:
    A binary tree can be created as follows:

        bt = BinaryTreeNode(1)
        bt.left = BinaryTreeNode(2)
        bt.right = BinaryTreeNode(3)

    An n-ary tree can be created as follows, where one can also get the depth of one of the children:

        nt = NaryTreeNode(1)
        c1 = NaryTreeNode(2)
        c2 = NaryTreeNode(3)
        c3 = NaryTreeNode(4)
        c1.parent = nt
        c2.parent = nt
        c3.parent = nt
        nt.leftmost_child = c1
        c1.right_sibling = c2
        c2.right_sibling = c3
        x = depth(c3)
"""


class BinaryTreeNode:
    """Represents a binary tree node.

    Attributes:
        data: The data stored in the node. This can be any type.
        left: A `BinaryTreeNode` representing the node that is to the left of this node in a tree.
        right: A `BinaryTreeNode` representing the node that is to the right of this node in a tree.
    """

    def __init__(self, data=None, left=None, right=None):
        """Initializes a `BinaryTreeNode` in `O(1)` time.

        Args:
            data: The data to be stored in the node. This can be of any type. By default, this is `None`.
            left: A `BinaryTreeNode` representing the node that is to the left of this node in a tree. By default this
                  is `None`.
            right: A `BinaryTreeNode` representing the node that is to the right of this node in a tree. By default this
                  is `None`.
        """
        self.data = data
        self.left = left
        self.right = right


class NaryTreeNode:
    """Represents a n-ary tree node.

    Attributes:
        data: The data stored in the node. This can be any type.
        parent: A `NaryTreeNode` representing the node that is the parent of this node in a tree.
        leftmost_child: A `NaryTreeNode` representing the node that is the leftmost child of this node in a tree.
        right_sibling: A `NaryTreeNode` representing the node that is the right sibling of this node in a tree.
    """

    def __init__(self, data=None, parent=None, leftmost_child=None, right_sibling=None):
        """Initializes a `NaryTreeNode` in `O(1)` time.

        Args:
            data: The data to be stored in the node. This can be of any type. By default, this is `None`.
            parent: A `NaryTreeNode` representing the node that is the parent of this node in a tree. By default, this
                    is `None`.
            leftmost_child: A `NaryTreeNode` representing the node that is the leftmost child of this node in a tree. By
                            default, this is `None`.
            right_sibling: A `NaryTreeNode` representing the node that is the right sibling of this node in a tree. By
                           default, this is `None`.
        """
        self.data = data
        self.parent = parent
        self.leftmost_child = leftmost_child
        self.right_sibling = right_sibling


def depth(node):
    """Computes the depth of a n-ary tree node.

    The depth of a node `v` is the number of edges on the path from the root to the node. One may assume this function
    runs in `O(h)` time where `h` is the height of tree the input node is contained in.

    Args:
        node: an `NaryTreeNode`.

    Returns:
        The integer depth of node in the tree its contained in.

    Raises:
        TypeError: If `node` is not an `NaryTreeNode`.
    """
    if not isinstance(node, NaryTreeNode):
        raise TypeError(f'Can only calculate depth of NnaryTreeNode not {type(node)}')
    levels = 0
    while node.parent is not None:
        node = node.parent
        levels += 1
    return levels



"""This module holds classes related to graphs.

This module contains the `Vertex`, `VertexAttributeError`, `Graph`, and `GraphVertexError` classes. `Vertex` represents
a graph vertex. `VertexAttributeError` is an error raised by `Vertex`. `Graph` represents a directed graph.
`GraphVertexError` is an error raised by `Graph`.

Examples:
    Initializing a `Graph` as well as adding colored vertices and edges:

        g = Graph()
        a = Vertex('a', color='red')
        b = Vertex('b', color='blue')
        g.add_vertex(a)
        g.add_vertex(b)
        g.add_edge(a, b)

    The following code will raise a `VertexAttributeError` because `'colour'` is not an attribute of the `Vertex` `a`:

        x = a['colour']
"""

from dalpy.sets import Set


class VertexAttributeError(Exception):
    """This class is used by `Vertex` to raise errors regarding invalid attributes."""

    def __init__(self, name, wrong_attribute, attributes):
        """Initializes a `VertexAttributeError`.

        Args:
            name: String name of `Vertex` that erroneous attribute is associated with.
            wrong_attribute: String attribute that does not exist in the `Vertex` referred to by `name`.
            attributes: The valid attributes in the `Vertex` referred to by `name` as a `list` of `str`.
        """
        super().__init__(
            f'vertex {name} does not have attribute {wrong_attribute}, available attributes: [{", ".join(a for a in attributes)}]')


class Vertex:
    """Represents a graph vertex.

    A `Vertex` has a name and a collection of customizable attributes. The name should be used to identify it in a
    `Graph`. `Vertex` objects are compared with `==` and via hash code based on their name. The attributes you can use
    in graph algorithms. For example, you could set colors or time steps as in DFS. One should assume all operations
    that can be performed on a `Vertex` are done in `O(1)` time.

    Examples:
        To initialize a `Vertex` with a name:

            v = Vertex('a')

        To initialize a `Vertex` with some starting attributes, use keyword arguments following the name:

            v = Vertex('a', color='red', time=1)

        To add additional attributes to an existing `Vertex` using `[]` with `=`:

            v['seen'] = False

        To update or view existing attributes, use the same idea:

            x = v['color']
            v['time'] = 3

        If you try to get the value of an attribute that does not exist, a `VertexAttributeError` will be raised (e.g.
        calling `print(v['colour'])`. If you try to add a new attribute to an existing `Vertex` where the attribute's
        type is not `str`, a `TypeError` will be raised (e.g. calling `v[1] = 'blue'`).
    """

    def __init__(self, name, **attributes):
        """Initializes a `Vertex`.

        Args:
            name: The name of the `Vertex`. Make sure to read the class docstring above regarding naming of `Vertex`
                  objects.
            **attributes: Optional keyword arguments specifying attributes you want the `Vertex` to start with. See
                          the class docstring above for some possible attributes you could add.
        """
        self.__name = name
        self.__attributes = attributes

    def get_name(self):
        """Returns the name of this `Vertex`."""
        return self.__name

    def __getitem__(self, attribute):
        if attribute not in self.__attributes:
            raise VertexAttributeError(self.__name, attribute, self.__attributes)
        return self.__attributes[attribute]

    def __setitem__(self, attribute, value):
        if attribute not in self.__attributes and not isinstance(attribute, str):
            raise TypeError(f'attribute {attribute} is not of type str, has type {type(attribute)}')
        self.__attributes[attribute] = value

    def __hash__(self):
        return hash(self.__name)

    def __eq__(self, other):
        if not isinstance(other, Vertex):
            return False
        return other.__name == self.__name

    def __iter__(self):
        # Disables the users' ability to do for a in v for an Vertex v. Without disabling this, the default Python
        # __iter__ will call __getitem__ with an int. We have specified __getitem__ only accepts strings.
        raise RuntimeError(
            'Not allowed to iterate over Vertex like an iterable. Must go through attributes individually with [].')


class GraphVertexError(Exception):
    """This class is used by `Graph` to raise errors regarding invalid vertices."""

    def __init__(self, vertex_name):
        """Initializes a `GraphVertexError` that will be raised associated with a particular `Vertex`.

        Args:
            vertex_name: The string name of the `Vertex` this `GraphVertexError` is being raised in association with.
        """
        super().__init__(f'graph does not have vertex with name {vertex_name}')


class GraphEdgeError(Exception):
    """This class is used by `Graph` to raise errors regarding invalid edges."""

    def __init__(self, source_name, dest_name):
        """Initializes a `GraphEdgeError` that will be raised associated with a particular `Vertex`.

        Args:
            source_name: The string name of the source `Vertex` of the edge this `GraphVertexError` is being raised in
                         association with.
            dest_name: The string name of the destination `Vertex` of the edge this `GraphVertexError` is being raised
                       in association with.
        """
        super().__init__(
            f'graph does not have an edge from a vertex with name {source_name} to a vertex with name {dest_name}')


class Graph:
    """Represents a directed graph.

    A `Graph` represents a directed graph object whose edges can be assigned weights. The `Vertex` objects in the
    `Graph` should have unique names. For more information on this, see the class docstring of `Vertex`.

    Examples:
        To initialize a `Graph`:

            g = Graph()

        In order to add edges to a `Graph`, one must first add the `Vertex` objects that will make up that edge.

            a = Vertex('a')
            b = Vertex('b')
            g.add_vertex(a)
            g.add_vertex(b)

        Now that `a` and `b` have been added to `g`, one can add an edge between them. Note that there is no edge
        object.

            g.add_edge(a, b)

        When adding an edge, one can specify a weight (by default it is `None`):

            g.add_edge(a, b, 1)

        One can get a `dalpy.sets.Set` of the adjacent edges of a Vertex:

            s = g.adj(a)
    """

    def __init__(self):
        """Initializes an empty `Graph` in `O(1)` time."""
        self.__adj_lists = dict()
        self.__edge_weights = dict()

    def add_vertex(self, vertex):
        """Adds a `Vertex` to this `Graph`.

        One may assume that this operation runs in `O(1)` time with respect to the number of vertices and edges in this
        `Graph`.

        Args:
            vertex: The `Vertex` to be added to this `Graph`.
        """
        self.__adj_lists[vertex] = list()

    def add_edge(self, source, dest, weight=None):
        """Adds an edge between 2 `Vertex` objects in this `Graph`.

        This creates an edge from the source `Vertex` to the destination `Vertex`. `Graph` is directed so the edge is
        in only one direction. That is, an edge from the destination `Vertex` to the source `Vertex` will not exist. One
        can additionally provide a weight for this edge. One may assume that this operation runs in `O(1)` time with
        respect to the number of vertices and edges in this `Graph`.

        Args:
            source: The source `Vertex`.
            dest: The destination `Vertex`.
            weight: The weight for this edge (floating point, integer). By default this is `None`.

        Raises:
            GraphVertexError: If `source` or `dest` is not in this `Graph`.
        """
        self.__check_edge_vertices(source, dest)
        self.__edge_weights[(source, dest)] = weight
        self.__adj_lists[source].append(dest)

    def adj(self, vertex):
        """Gets the `Vertex` objects that are adjacent to a `Vertex`.

        This gets a `dalpy.sets.Set` of the `Vertex` objects that are adjacent to the vertex. Since
        `dalpy.sets.Set` objects preserve insertion order (see `dalpy.sets.Set` documentation), the
        `dalpy.sets.Set` will be ordered according to the order in which edges starting from the input `Vertex`
        were created. One should assume that this method runs in `O(n)` time where `n` is the number of edges going out
        of the input `Vertex`.

        Args:
            vertex: A `Vertex`.

        Returns:
            A `dalpy.sets.Set` containing the vertices adjacent to `vertex`.

        Raises:
            GraphVertexError: If `vertex` is not in this `Graph`.

        Examples:
            To illustrate the nature of the `dalpy.sets.Set` returned by this method, first set up a `Graph` and
            add some vertices and edges:

                g = Graph()
                a = Vertex('a')
                b = Vertex('b')
                c = Vertex('c')
                g.add_vertex(a)
                g.add_vertex(b)
                g.add_vertex(c)
                g.add_edge(a, b)
                g.add_edge(a, c)

            The `dalpy.sets.Set` returned by `g.adj(a)` will always have `b` preceding `c` since the edge from `a`
            to `b` was created before the edge from `a` to `c`.
        """
        if vertex not in self.__adj_lists:
            raise GraphVertexError(vertex.get_name())
        adjacent = Set()
        for u in self.__adj_lists[vertex]:
            adjacent.add(u)
        return adjacent

    def weight(self, source, dest):
        """Gets the weight of an edge defined by two vertices.

        One should assume that this method runs in `O(1)` time with respect to the number of vertices and edges in this
        `Graph`.

        Args:
            source: The source `Vertex` of the edge in question.
            dest: The destination `Vertex` of the edge in question.

        Returns:
            The integer or floating point weight associated with the edge from `source` to `dest`. This will be `None`
            if no weight was specified when the edge was created.

        Raises:
            GraphVertexError: If `source` or `dest` is not in this `Graph`.
            GraphEdgeError: If an edge from `source` to `dest` does not exist.
        """
        self.__check_edge_vertices(source, dest)
        if (source, dest) not in self.__edge_weights:
            raise GraphEdgeError(source, dest)
        return self.__edge_weights[(source, dest)]

    def vertices(self):
        """Gets the vertices in this `Graph`.

        One should assume that this method runs in `O(V)` time where `V` is the number of vertices added to this
        `Graph`.

        Returns:
            A `dalpy.sets.Set` containing the `Vertex` objects in this `Graph`. The order of the vertices in this
            `dalpy.sets.Set` will always be the order in which the vertices were added to this `Graph` via
            `add_vertex`.
        """
        s = Set()
        for vertex in self.__adj_lists:
            s.add(vertex)
        return s

    def __check_edge_vertices(self, source, dest):
        if source not in self.__adj_lists:
            raise GraphVertexError(source.get_name())
        if dest not in self.__adj_lists:
            raise GraphVertexError(dest.get_name())



"""Module that holds classes related to sets.

This module contains the `Set` class. `Set` represents a set.

Examples:
    Creating a `Set`, adding elements, and checking membership:

        s = Set()
        s.add(1)
        s.add(2)
        s.add(3)
        if 1 in s:
            print("1")

    Removing elements from a `Set`:

        t = Set()
        t.add(3)
        t.add(4)
        s.difference(t)
"""


class Set:
    """Represents a set.

    This class represents a set that preserves insertion order. This allows for it to be used in cases where the order
    of the set's contents must be deterministic.

    Examples:
        To initialize a `Set`:

            s = Set()

        To add elements to `s`:

            s.add(1)
            s.add(2)

        To check if `s` contains an element:

            if 1 in s:
                # Do something

        To remove a Set of elements from `s`, make use of difference:

            t = Set()
            t.add(2)
            t.add(3)
            s.difference(t)

        To remove one element from `s`, make use of the ability to create a singleton set combined with difference:

            s.difference(Set(1))

        To iterate over a `Set`:

            for e in s:
                # Do something with e
    """

    def __init__(self, singleton_elem=None):
        """Initializes a `Set` in `O(1)` time.

        Args:
            singleton_elem: To initialize the `Set` to be a singleton `Set`, pass a non-`None` value as this parameter
                            (of any type). By default, this is `None`, in which case the `Set` is initialized to be
                            empty.
        """

        # https://docs.python.org/3/library/stdtypes.html#dict
        # Dictionaries preserve insertion order. Note that updating a key does not affect the order.
        # Keys added after deletion are inserted at the end.
        self.__set = dict()
        if singleton_elem is not None:
            self.__set[singleton_elem] = None

    def add(self, element):
        """Adds an element to the Set.

        Args:
            element: An element to add. If element is already in the `Set`, then it is not added again.

        Raises:
            ValueError: If element is `None`. This is to preserve the interpretation of `None` when initializing a
                        `Set`.
        """
        if element is None:
            raise ValueError('Cannot add None to a Set.')
        self.__set[element] = None

    def difference(self, other_set):
        """Performs a set difference operation on this `Set`.

        This method removes all the elements from this `Set` that occur in another set. A user should assume this runs
        in `O(n)` time where `n` is the size of the other `Set`.

        Args:
            other_set: Another `Set` specifying the elements to be removed from this `Set`. This `Set` is unaffected by
                       this method.

        Raises:
            TypeError: If `other_set` is not a `Set`.
        """
        if not isinstance(other_set, Set):
            raise TypeError(f'can only perform set difference between Sets')
        for k in other_set.__set:
            if k in self.__set:
                self.__set.pop(k)

    def is_empty(self):
        """Returns `True` if this `Set` is empty, `False` otherwise in `O(1)` time w/r/t the size of this `Set`."""
        return len(self.__set) == 0

    def size(self):
        """Returns the integer number of elements in this `Set` in `O(1)` time w/r/t the size of this `Set`."""
        return len(self.__set)

    def __contains__(self, element):
        return element in self.__set

    def __iter__(self):
        yield from iter(self.__set)



"""This module provides factory functions for creating and modifying DALPy objects.

The functions contained in this module are meant to be used when writing test cases for problems implemented in
DALPy.
"""

from dalpy.stacks import Stack
from dalpy.arrays import Array
from dalpy.queues import Queue


def make_queue(ls):
    """Creates a `Queue` from a `list` of elements, enqueueing from left to right.

    Args:
        ls: A `list` of elements to be enqueued into the `Queue`.
    """
    queue = Queue()
    for x in ls:
        queue.enqueue(x)
    return queue


def make_stack(ls):
    """Creates a `Stack` from a `list` of elements, pushing from left to right.

    Args:
        ls: A `list` of elements to be enqueued into the `Stack`.
    """
    s = Stack()
    for e in ls:
        s.push(e)
    return s


def copy_stack(s):
    """Creates a copy of a `Stack`.

    Args:
        s: The `Stack` to copy.
    """
    buf = Stack()
    while not s.is_empty():
        nxt = s.pop()
        buf.push(nxt)
        buf.push(nxt)
    cpy = Stack()
    while not buf.is_empty():
        s.push(buf.pop())
        cpy.push(buf.pop())
    return cpy


def make_array(ls):
    """Makes an `Array` from a `list` of elements.

    Args:
        ls: The `list` of elements to insert into a new `Array`.
    """
    a = Array(len(ls))
    for i, e in enumerate(ls):
        a[i] = e
    return a


def nary_add_children(root, children):
    """Adds `NaryTreeNode`s to be the children of another `NaryTreeNode`.

    Args:
        root: an `NaryTreeNode` to which we will be adding children. This will be modified in place by this function.
        children: a `list` of `NaryTreeNode`s that are the children to be added. These nodes will be added as the
                  children of `root` in the order they occur in this `list`. `children[0]` will be the leftmost child of
                  `root`, `children[1]` will be the right sibling of `children[0]` and so on.
    """

    if len(children) == 0:
        return
    root.leftmost_child = children[0]
    for i, child in enumerate(children):
        child.parent = root
        if i < len(children) - 1:
            child.right_sibling = children[i + 1]



"""This module provides utilities related to unit testing.

This module contains the `build_and_run_watched_suite`, `assert_array_equals`, `behavior_test`, `generic_test`,
`dalpy_equals` and the `dalpy_to_string` functions, as well as `UnexpectedReturnWarning`.
"""
import copy
import inspect
import math
import traceback
import unittest
import warnings
from multiprocessing import Process

from dalpy.arrays import Array, Array2D
from dalpy.factory_utils import copy_stack
from dalpy.graphs import Graph, Vertex
from dalpy.linked_lists import SinglyLinkedListNode
from dalpy.queues import Queue
from dalpy.sets import Set
from dalpy.stacks import Stack
from dalpy.trees import BinaryTreeNode, NaryTreeNode


def build_and_run_watched_suite(cases, timeout=None, show_tb=False, grading_file=None, warning_filter="once"):
    """Runs a set of test cases, ensuring that they do not run longer than `timeout` seconds. Optionally,
    writes comma-separated test results to a file.

    Args:
        cases: A list of TestCases to be run.
        timeout: Number of seconds to allow each test case to run for.
        show_tb: Boolean toggle for stack trace.
        grading_file: Output file path to store comma-separated test results.
        warning_filter: A `warnings.simplefilter` action. Default value ensures that warnings are only displayed once.
                        Choose `"ignore"` to suppress warnings.

    If `grading_file` is not specified, the test logs will be dumped to console.
    """

    def _warning(
            message,
            category,
            filename,
            lineno,
            file=None,
            line=None):
        print(f'{__bcolors.WARNING}{category.__name__}: {message}{__bcolors.ENDC}')

    warnings.showwarning = _warning
    warnings.simplefilter(warning_filter, UnexpectedReturnWarning)
    warnings.simplefilter("once", DeprecationWarning)
    watcher = _Watcher(show_tb)
    suite = unittest.TestSuite()
    for case in cases:
        tests = unittest.defaultTestLoader.loadTestsFromTestCase(case)
        suite.addTests(tests)
    for test in suite:
        __run_timed_test(test, watcher, timeout)
    if grading_file is not None:
        watcher.print_columns(grading_file)
    else:
        watcher.print_log()


def assert_array_equals(expected, actual, msg=None):
    """Asserts that two `dalpy.arrays.Array`s are equal, displaying a custom message if specified.

    Args:
        expected: The expected `dalpy.arrays.Array`.
        actual: The actual `dalpy.arrays.Array`.
        msg: The message to display on `AssertionError`, if not specified, then a default message is displayed.

    Raises:
        AssertionError: If `expected` != `actual`.
    """
    assert isinstance(actual, Array)
    assert expected.length() == actual.length(), f"Expected Array length = {expected.length()}, Actual Array length = {actual.length()}" if msg is None else msg
    for i in range(expected.length()):
        assert expected[i] == actual[
            i], f"Expected {expected[i]} at index {i}, Actual = {actual[i]}" if msg is None else msg


def behavior_test(behavior, objects):
    """Test the behavior of an object.

    Args:
        behavior: a `list` of `tuple`s of the form `(RESULT, METHOD, PARAMETERS)`.
        objects: a `list` of objects who's parameters are being called.

    Raises:
        AssertionError: If `METHOD(PARAMETERS) != RESULT`.

    For each tuple in behavior this test asserts that `METHOD(PARAMETERS) = RESULT`.

    In each `tuple` `METHOD` should an uncalled `callable`, for example:
    >>> stack = Stack()
    >>> uncalled_callable = stack.pop

    Notes:
    - If `METHOD` requires multiple parameters, then `PARAMETERS` can be passed as a `tuple`.
    - If `METHOD` has no required return, then `RESULT` can be omitted in favor of `(METHOD, PARAMETERS)`.
    - If `METHOD` has no parameters, then `PARAMETERS` can be omitted in favor of `(RESULT, METHOD)`.

    Example:

    >>> stack = Stack()
    >>> behavior = [ (stack.push, 1), (1, stack.pop) ]

    The objects parameter is the object who's behavior is being tested, which will be used for the test log.
    If multiple objects are being tested, pass a tuple of objects.
    """
    msg = f'Behavior:\ninit {", ".join(type(obj).__name__ for obj in objects) if isinstance(objects, list) else type(objects).__name__}\n'
    passed = True
    expected, method, params, result = None, None, None, None
    if not isinstance(behavior, list): behavior = [behavior]
    try:
        for event in behavior:
            if len(event) == 2 and callable(event[0]) and not isinstance(event[0], type): event = (None,) + event
            expected = event[0]
            method = event[1]
            if len(event) > 2:
                params = (event[2],)
                result = method(*params)
                msg += f'{__method_to_string((method,) + params)} {dalpy_to_string(result)}'
            else:
                result = method()
                msg += f'{__method_to_string(method)} {dalpy_to_string(result)}'
                params = None
            if dalpy_equals(result, expected):
                msg += ' ✓\n'
                continue
            msg += f' ✗\nexpected {dalpy_to_string(expected)}'
            passed = False
            break
    except Exception as e:
        # If expected is an exception then check that exception thrown matches expected exception
        msg += f'{__method_to_string((method,) + params if params is not None else method)} {dalpy_to_string(result)} ✗\n'
        if type(expected) == type and isinstance(e, expected): return
        # error_message = e.args[0] if len(e.args) > 0 else e.with_traceback
        assert False, f'{msg}Unexpected error: {type(e).__name__}'
    assert passed, msg


def run_generic_test(params, expected, method, custom_comparator=None, in_place=False, enforce_no_mod=False,
                     init_params=None,
                     init_expected=None, params_to_string=None, expected_to_string=None, output_to_string=None):
    """Test the output of a function.

    Warnings:
        Deprecated in 1.1.0, to be removed. Use the generic_test function instead.

    Args:
        params: Parameters to be passed into the function being tested. This argument can either be a single parameter,
                or a list of parameters.
        expected: Expected return value of tested function with parameters specified by params.
        method: Function being tested. Must be a `callable`.
        custom_comparator: Function for determining if method output equals expected. Must be a `callable`.
        in_place: `True` if `expected` should be compared against `params`.
        enforce_no_mod: `bool` or a `list` of `bool` indicating which args should not be modified. Default `False`
                        allows modification of all args.
        init_params: Function for initializing parameters. Must be a `callable`.
        init_expected: Function for initializing expected output. Must be a `callable`.
        params_to_string: Function for displaying the parameters. Must be a `callable`.
        expected_to_string: Function for displaying the expected output. Must be a `callable`.
        output_to_string: Function for displaying the actual output. Must be a `callable`.

    Raises:
        AssertionError: If the test fails.
        UnexpectedReturnWarning: If `in_place` is set to `True` but `method` still returns a value.
        DeprecationWarning: If used in version >= 1.1.0.

    If `expected` is an `Exception`, the test will assert that the function tested on the given parameters throws the
    expected `Exception`. If no custom `to_string`s are specified, the `dalpy_to_string` method will be used for
    displaying parameters, input and output.
    """
    warnings.warn("run_generic_test is deprecated after version 1.1.0, use generic_test instead.", DeprecationWarning,
                  stacklevel=2)
    params = init_params(params) if init_params is not None else params
    expected = init_expected(expected) if init_expected is not None else expected
    generic_test(params, expected, method, custom_comparator=custom_comparator, in_place=in_place,
                 enforce_no_mod=enforce_no_mod,
                 params_to_string=params_to_string, expected_to_string=expected_to_string,
                 output_to_string=output_to_string)


def generic_test(params, expected, method, custom_comparator=None, in_place=False, enforce_no_mod=False,
                 params_to_string=None, expected_to_string=None, output_to_string=None):
    """Test the output of a function.

    Args:
        params: Parameters to be passed into the function being tested. This argument can either be a single parameter,
                or a list of parameters.
        expected: Expected return value of tested function with parameters specified by params. If `expected` is an
                  `Exception`, the test will assert that the function tested on the given parameters throws the expected
                  `Exception`.
        method: Function being tested. Must be a `callable`.
        custom_comparator: Function for determining if method output equals expected. Must be a `callable`. Default
                           `None` which means that `dalpy_equals` will be used.
        in_place: `True` if `expected` should be compared against `params`. By default this is `False`.
        enforce_no_mod: `bool` or a `list` of `bool` indicating which args should not be modified. Default `False`
                        allows modification of all args.
        params_to_string: Function for displaying the parameters. Must be a `callable`. Default `None` which means that
                          `dalpy_to_string` will be used instead.
        expected_to_string: Function for displaying the expected output. Must be a `callable`. Default `None` which
                            means that `dalpy_to_string` will be used instead.
        output_to_string: Function for displaying the actual output. Must be a `callable`. Default `None` which means
                          that `dalpy_to_string` will be used instead.

    Raises:
        AssertionError: If the test fails.
        UnexpectedReturnWarning: If `in_place` is set to `True` but `method` still returns a value.
    """
    msg = f"Input: {dalpy_to_string(params) if params_to_string is None else params_to_string(params)}\nExpected: {dalpy_to_string(expected) if expected_to_string is None else expected_to_string(expected)}\n"
    params_copy = copy.deepcopy(params) if isinstance(params, list) else [copy.deepcopy(params)]
    passed = True
    try:
        result = method(*params) if isinstance(params, list) else method(params)
        if in_place:
            if result is not None:
                warnings.warn("A function that is meant to modify its argument(s) returned a non-None value.",
                              UnexpectedReturnWarning, stacklevel=2)
            result = params
        result_string = output_to_string(result) if output_to_string is not None else dalpy_to_string(result)
        if custom_comparator is None:
            if not dalpy_equals(expected, result):
                msg = f"{msg}Output: {result_string}"
                passed = False
        elif not custom_comparator(expected, result):
            msg = f"{msg}Output: {result_string}"
            passed = False
    except Exception as e:
        # If expected is an exception then check that exception thrown matches expected exception
        if type(expected) == type and isinstance(e, expected): return
        error_message = e.args[0] if len(e.args) > 0 else e.with_traceback
        assert False, f"{msg}Output: {error_message}"
    assert passed, msg
    enforce_no_mod = [enforce_no_mod] * len(params_copy) if isinstance(enforce_no_mod, bool) else enforce_no_mod
    modified_params_string = dalpy_to_string(params) if params_to_string is None else params_to_string(params)
    if not isinstance(params, list): params = [params]
    for i, no_mod in enumerate(enforce_no_mod):
        if no_mod:
            assert dalpy_equals(params_copy[i], params[
                i]), f"{msg}Output: The {str(i + 1) + __append_int(i + 1)} input argument should not have been modified.\nArguments: {modified_params_string}"


def dalpy_equals(first, second):
    """Tests equality between two objects. If the objects are from the DALPy, they are compared using their own
    custom comparator.

    `dalpy_equals` supports equality for the following objects: `dalpy.arrays.Array`, `dalpy.arrays.Array2D`,
    `dalpy.queues.Queue`, `dalpy.stacks.Stack`, `dalpy.sets.Set`,
    `dalpy.linked_lists.SinglyLinkedListNode`. For `dalpy.linked_lists.SinglyLinkedListNode`, checks that all
    nodes next of the passed `dalpy.linked_lists.SinglyLinkedListNode`s are the same. For instances of `float`s,
    `math.isclose` is used for comparison.

    Args:
        first: The first element to be tested.
        second: The second element to be tested

    Returns:
        `True` if `first = second` otherwise `False`.
    """
    if isinstance(first, Array) and isinstance(second, Array):
        return __array_equals(first, second)
    if isinstance(first, Array2D) and isinstance(second, Array2D):
        return __array2d_equals(first, second)
    if isinstance(first, Queue) and isinstance(second, Queue):
        return __queue_equals(first, second)
    if isinstance(first, Stack) and isinstance(second, Stack):
        return __stack_equals(first, second)
    if isinstance(first, Set) and isinstance(second, Set):
        return __set_equals(first, second)
    if isinstance(first, SinglyLinkedListNode) and isinstance(second, SinglyLinkedListNode):
        return __singly_linked_list_equals(first, second)
    if isinstance(first, float) and isinstance(second, float):
        return math.isclose(first, second)
    return first == second


def dalpy_to_string(obj):
    """Generates a string representation of a DALPy object if passed object is from DALPy, otherwise calls
    native str method.

    dalpy_to_string supports the following objects: `dalpy.arrays.Array`, `dalpy.arrays.Array2D`,
    `dalpy.queues.Queue`, `dalpy.stacks.Stack`, `dalpy.sets.Set`,
    `dalpy.linked_lists.SinglyLinkedListNode`, `dalpy.trees.BinaryTreeNode`, `dalpy.trees.NaryTreeNode`,
    `dalpy.graphs.Vertex`, and `dalpy.graphs.Graph`.

    Returns:
        string representation of `obj`.

    Args:
        obj: The object to convert to string
    """
    if isinstance(obj, list):
        return "[" + ", ".join(dalpy_to_string(elem) for elem in obj) + "]"
    if isinstance(obj, Array):
        return __array_to_string(obj)
    if isinstance(obj, Array2D):
        return __array2d_to_string(obj)
    if isinstance(obj, Queue):
        return __queue_to_string(obj)
    if isinstance(obj, Stack):
        return __stack_to_string(obj)
    if isinstance(obj, Set):
        return __set_to_string(obj)
    if isinstance(obj, SinglyLinkedListNode):
        return __singly_linked_list_to_string(obj)
    if isinstance(obj, BinaryTreeNode):
        return __binary_tree_to_string(obj)
    if isinstance(obj, NaryTreeNode):
        return __nary_tree_to_string(obj)
    if isinstance(obj, Vertex):
        return __vertex_to_string(obj)
    if isinstance(obj, Graph):
        return __graph_to_string(obj)
    try:
        return str(obj)
    except:
        return obj


class UnexpectedReturnWarning(Warning):
    """A `Warning` subclass for instances where functions are expected to modify their arguments but return values instead."""
    pass


class _TestTimeoutError(Exception):
    def __init__(self, timeout):
        super().__init__(f'Test timed out after {timeout}s.')


class _Watcher(unittest.TestResult):

    def __init__(self, show_tb):
        super().__init__()
        self.test_ids = list()
        self.results = list()
        self.details = list()
        self.show_tb = show_tb

    @staticmethod
    def parse_id(test_id):
        return test_id[test_id.index('.') + 1:]

    @staticmethod
    def get_description(test):
        if test.shortDescription() is None: return '\n'
        lines = test._testMethodDoc.split('\n')
        return "\n".join(line.strip() for line in lines) + "Output:\t"

    def addSuccess(self, test) -> None:
        self.test_ids.append(_Watcher.parse_id(test.id()))
        self.results.append(1)

    def addFailure(self, test, err) -> None:
        self.test_ids.append(_Watcher.parse_id(test.id()))
        self.results.append(0)
        tb_str = ''
        if self.show_tb:
            tb_str = '\n' + ''.join(traceback.format_tb(err[2]))
        self.details.append((_Watcher.get_description(test), str(err[1]) + tb_str, _Watcher.parse_id(test.id())))

    def addError(self, test, err):
        self.test_ids.append(_Watcher.parse_id(test.id()))
        self.results.append(0)
        tb_str = ''
        if self.show_tb:
            tb_str = '\n' + ''.join(traceback.format_tb(err[2]))
        self.details.append(
            (_Watcher.get_description(test), err[0], str(err[1]) + tb_str, _Watcher.parse_id(test.id())))

    # Note The order in which the various tests will be run is determined by sorting the test method names with respect
    # to the built-in ordering for strings.
    def print_columns(self, fp):
        with open(fp, mode='w+', encoding='utf-8') as f:
            f.write(",".join(self.test_ids))
            f.write('\n')
            f.write(",".join(str(e) for e in self.results))

    def print_log(self):
        log = list()
        for detail in self.details:
            if len(detail) == 3:
                log.append(f'{detail[2].split("Test.")[0]} test failed.\n{detail[0]}{detail[1]}')
            else:
                log.append(f'{detail[0]} raised {detail[1]}.\nMessage: {detail[2]}')
        print("\n" + ("\n" + "-" * 40 + "\n").join(
            log) + f"\n{'=' * 40}\n{sum(self.results)}/{len(self.results)} tests passed.\n")


def __run_timed_test(test, watcher, timeout):
    if timeout is not None:
        p = Process(target=test.run)
        p.start()
        p.join(timeout=timeout)
        if p.is_alive():
            p.terminate()
            watcher.addError(test, (_TestTimeoutError, _TestTimeoutError(timeout)))
            return
    test.run(watcher)


def __array_to_string(array):
    out = "["
    for i in range(array.length()):
        out += f'{dalpy_to_string(array[i])}, '
    return out[:-2] + "]" if array.length() > 0 else out + "]"


def __array2d_to_string(array):
    out = "["
    for i in range(array.rows()):
        out += "["
        for j in range(array.columns()):
            out += f'{dalpy_to_string(array[(i, j)])}, '
        out = out[:-2] + "]\n "
    return out[:-2] + "]"


def __queue_to_string(queue):
    out = []
    for _ in range(queue.size()):
        next = queue.dequeue()
        out.append(dalpy_to_string(next))
        queue.enqueue(next)
    return "[" + ", ".join(out) + "]"


def __stack_to_string(stack):
    out = []
    temp_stack = Stack()
    for _ in range(stack.size()):
        next = stack.pop()
        out.insert(0, dalpy_to_string(next))
        temp_stack.push(next)
    while not temp_stack.is_empty():
        stack.push(temp_stack.pop())
    return "[" + ", ".join(out) + "]"


def __set_to_string(s):
    out = []
    for elem in s:
        out.append(dalpy_to_string(elem))
    return "{" + ", ".join(out) + "}"


def __singly_linked_list_to_string(head):
    out = list()
    seen = set()
    while head is not None:
        if head in seen:
            out.append("cycle")
            break
        seen.add(head)
        out.append(dalpy_to_string(head.data))
        head = head.next
    return "➔ ".join(out)


def __strip_trailing_nones(ls):
    while len(ls) > 0 and ls[-1] is None:
        ls.pop()


def __binary_tree_to_string(root):
    # https://leetcode.com/problems/balanced-binary-tree/
    if root is None:
        return ""

    out_buf = list()
    q = list()
    q.append(root)
    all_none_level = False
    while not all_none_level:
        k = len(q)
        all_none_level = True
        for _ in range(k):
            curr = q.pop(0)
            if curr is not None:
                q.append(curr.left)
                q.append(curr.right)
                out_buf.append(dalpy_to_string(curr.data))
                all_none_level = False
            else:
                out_buf.append(None)
    __strip_trailing_nones(out_buf)
    return str(out_buf)


def __vertex_to_string(vertex):
    return vertex.get_name()


def __graph_to_string(graph):
    contents = list()
    for vertex in graph.vertices():
        edges = list()
        for dest in graph.adj(vertex):
            edge_str = f'{dest.get_name()}'
            weight = graph.weight(vertex, dest)
            if weight is not None:
                edge_str += f' <{weight}>'
            edges.append(edge_str)
        edges = ', '.join(edges)
        contents.append(f'{vertex.get_name()}: {edges}')
    contents = '\n'.join(contents)
    return contents


def __nary_tree_to_string(root):
    # https://leetcode.com/problems/n-ary-tree-preorder-traversal/
    if root is None:
        return ""
    out = list()
    q = [root, root.right_sibling]
    while len(q) > 0:
        k = len(q)
        for _ in range(k):
            curr = q.pop(0)
            if curr is None:
                out.append(None)
            else:
                out.append(dalpy_to_string(curr.data))
                lm_child = curr.leftmost_child
                q.append(lm_child)
                while lm_child is not None:
                    lm_child = lm_child.right_sibling
                    q.append(lm_child)
    __strip_trailing_nones(out)
    return str(out)


def __array_equals(expected, actual):
    if expected.length() != actual.length(): return False
    for i in range(expected.length()):
        if expected[i] != actual[i]: return False
    return True


def __array2d_equals(expected, actual):
    if expected.rows() != actual.rows() or expected.columns() != actual.columns(): return False
    for i in range(expected.rows()):
        for j in range(expected.columns()):
            if expected[i, j] != actual[i, j]: return False
    return True


def __queue_equals(expected, actual):
    for _ in range(expected.size()):
        expected_elem = expected.dequeue()
        actual_elem = actual.dequeue()
        if expected_elem != actual_elem: return False
        expected.enqueue(expected_elem)
        actual.enqueue(actual_elem)
    return expected.size() == actual.size()


def __stack_equals(expected, actual):
    if expected.size() != actual.size(): return False
    expected_cpy = copy_stack(expected)
    actual_cpy = copy_stack(actual)
    i = 0
    while not expected_cpy.is_empty():
        e = expected_cpy.pop()
        a = actual_cpy.pop()
        if e != a: return False
        i += 1
    return True


def __set_equals(expected, actual):
    expected_set = set()
    actual_set = set()
    for elem in expected:
        expected_set.add(elem)
    for elem in actual:
        actual_set.add(elem)
    return expected_set == actual_set


def __singly_linked_list_equals(expected, actual):
    seen = set()
    while (expected is not None and actual is not None):
        if actual in seen: return False
        seen.add(actual)
        if expected != actual: return False
        expected = expected.next
        actual = actual.next
    return expected is None and actual is None


def __method_to_string(method):
    new_line = "\n"
    if isinstance(method, tuple):
        return f'({str(inspect.getsourcelines(method[0])[0][0]).strip(new_line).strip().split()[1].replace("(self,", "")}, {", ".join(str(param) for param in method[1:])})'
    return str(inspect.getsourcelines(method)[0][0]).strip("\n").strip().split()[1].replace('(self):', '()')


def __append_int(num):
    if num > 9:
        secondToLastDigit = str(num)[-2]
        if secondToLastDigit == '1':
            return 'th'
    lastDigit = num % 10
    if (lastDigit == 1):
        return 'st'
    elif (lastDigit == 2):
        return 'nd'
    elif (lastDigit == 3):
        return 'rd'
    else:
        return 'th'


class __bcolors:
    HEADER = '\033[95m'
    WARNING = '\033[93m'
    ENDC = '\033[0m'



"""Module that holds classes related to linked lists.

This module contains the ``SinglyLinkedListNode` and `DoublyLinkedListNode` classes. `SinglyLinkedListNode` can be used
to recursively define a singly linked list. `DoublyLinkedListNode` can be used to recursively define a doubly linked
list. Problems or algorithms that operate on linked lists should take an object of either of these classes.

Examples:
    A singly linked list can be created as follows:

        s = SinglyLinkedListNode(1)
        s.next = SinglyLinkedListNode(2)
        s.next.next = SinglyLinkedListNode(3)

    A doubly linked list can be created as follows:

        d = DoublyLinkedListNode(1)
        d.next = DoublyLinkedListNode(2)
        d.next.prev = d
        d.next.next = DoublyLinkedListNode(3)
        d.next.next.prev = d.next
"""


class SinglyLinkedListNode:
    """Represents a singly linked list node.

    Attributes:
        data: The data stored in the node. This can be any type.
        next: A `SinglyLinkedListNode` representing the node that follows it this node in a singly linked list.
    """

    def __init__(self, data=None, next_node=None):
        """Initializes a `SinglyLinkedListNode` in `O(1)` time.

        Args:
            data: The data to be stored in the node. This can be of any type. By default, this is `None`.
            next_node: The `SinglyLinkedListNode` to follow this `SinglyLinkedListNode`. By default, this is `None`.
        """
        self.data = data
        self.next = next_node


class DoublyLinkedListNode:
    """Represents a doubly linked list node.

    This class represents a doubly linked list node.

    Attributes:
        data: The data stored in the node. This can be of any type.
        next: A `DoublyLinkedListNode` representing the node that follows this node in a doubly linked list.
        prev: A `DoublyLinkedListNode` representing the node that precedes this node in a doubly linked list.
    """

    def __init__(self, data=None, next_node=None, prev_node=None):
        """Initializes a doubly linked list node in `O(1)` time.

        Args:
            data: The data to be stored in the node. This can be of any type. By default, this is `None`.
            next_node: The `DoublyLinkedListNode` to follow this `DoublyLinkedListNode`. By default, this is `None`.
            prev_node: The `DoublyLinkedListNode` to precede this `DoublyLinkedListNode`. By default, this is `None`.
        """
        self.data = data
        self.prev = prev_node
        self.next = next_node




"""Module that holds classes and functions related to 1 and 2 dimensional arrays.

This module contains the `Array` and `Array2D` classes and the `sort` function. `Array` represents a 1D array. `Array2D`
represents a 2D array. `sort` can be used to return a sorted copy of an `Array`.

Examples:
    Creating an `Array`, setting its elements, and getting a sorted copy:

        a = Array(3)
        a[0] = 1
        a[1] = 4
        a[2] = 2
        b = sort(a)

    Creating an `Array2D` and setting its elements:

        c = Array2D(3,2)
        for i in range(3):
            for j in range(2):
                c[i,j] = i + j
"""


class Array:
    """Represents a 1D array.

    This class represents a 1D fixed length array. It should be used like a Java array. That is, you cannot increase
    its length, slice it, or index it with negative numbers like a Python `list`. One cannot iterate over an `Array`
    like a Python `list` either. For example, `for e in a` will raise a `RuntimeError` for an `Array` `a`. A user should
    assume that indexing into an `Array` (either for getting or setting elements) is done in `O(1)` time with respect to
    the length of the `Array`.

    Examples:
        After constructing an `Array`, one can get an element at a particular index using square brackets `[]` as
        follows:

            a = Array(2)
            b = a[0]

        Here, `b` will store the first element in `a`.

        Similarly, one can set an element of an `Array` using a combination of `[]` and the assignment operator `=`

            a = Array(2)
            a[0] = 1

        Here, the first element of `a` is set to 1.

        Note that an `IndexError` will be raised in the event one supplies an index in `[]` that is negative or greater
        than or equal to the length of the `Array`.
    """

    def __init__(self, length):
        """Initializes an `Array` to be a particular length.

        Each of the length elements of this `Array` will be set to `None`. One should assume this runs in `O(length)`
        time.

        Args:
            length: An integer specifying the length of the `Array`.

        Raises:
            ValueError: If `length < 0`.
        """
        if length < 0:
            raise ValueError(f'length ({length}) can\'t be < 0')
        self.__buf = [None] * length

    def length(self):
        """Returns the integer length of this `Array` in `O(1)` time with respect to the length of the `Array`."""
        return len(self.__buf)

    def sort(self):
        """Sorts the elements of this `Array` in place.

        Note that this modifies this `Array`, it does not return a sorted copy. The user must make sure that the
        elements of the `Array` can be compared. For example, an empty `Array` cannot be sorted as `NoneType` (`None`)
        elements cannot be compared. A user may assume that this method runs in `O(n * log(n))` time where `n` is the
        length of the `Array`.
        """
        self.__buf.sort()

    def __getitem__(self, index):
        self.__check_index(index)
        return self.__buf[index]

    def __setitem__(self, index, value):
        self.__check_index(index)
        self.__buf[index] = value

    def __iter__(self):
        # Disables the users' ability to do for e in a for an Array a. Without disabling this, the default Python
        # __iter__ will call __getitem__ with an int.
        raise RuntimeError('Not allowed to iterate over Array like an iterable. Must go through `[]` with indices.')

    def __check_index(self, index):
        if index < 0 or index >= len(self.__buf):
            raise IndexError(f'index ({index}) can\'t be < 0 or >= length ({len(self.__buf)})')


class Array2D:
    """Represents a 2D Array.

    This class represents a 2D array with a fixed number of rows and columns. It should be used like a Java
    two-dimensional array. That is, you cannot increase its dimensions, slice it, or index it with negative numbers like
    a Python list. A user should assume that indexing into an `Array2D` (either for getting or setting) is done in
    `O(1)` time with respect to the dimensions of the `Array2D`.

    Examples:
        After constructing an `Array2D`, one can get an element at a particular (row, column) index using square
        brackets `[]` as follows:

            a = Array2D(2,2)
            b = a[0,1]

        Here, `b` will store the element in the second column of the first row of `a`.

        Similarly, one can set an element of an `Array2D` using a combination of `[]` and the assignment operator `=`

            a = Array2D(2,2)
            a[0,1] = 2

        Here, the element in the second column of the first of `a` is set to 2.

        Note that an `IndexError` will be raised in the event one supplies an index in `[]` that (i) has a row that is
        negative or greater than or equal to the number of rows of the `Array2D` or (ii) a column that is negative or
        greater than or equal to the number of columns of the `Array2D`.
    """

    def __init__(self, num_rows, num_columns):
        """Initializes an `Array2D` to be a particular number of rows and columns.

        Each of the `num_rows * num_columns` elements of this `Array` will be set to `None`. A user may assume that this
        method runs in `O(num_rows * num_columns)` time.

        Args:
            num_rows: An integer specifying the number of rows this `Array2D` should have.
            num_columns: An integer specifying the number of columns each of the rows of this `Array2D` should have.

        Raises:
            ValueError: If `num_rows <= 0` or `num_columns <= 0`.
        """
        if num_rows <= 0 or num_columns <= 0:
            raise ValueError(f'both num_rows ({num_rows}) and num_cols ({num_columns}) must be > 0')
        self.__buf = list()
        for _ in range(num_rows):
            self.__buf.append([None] * num_columns)

    def rows(self):
        """Gets the number of rows in this `Array2D`.

        A user should assume this runs in `O(1)` time with respect to the dimensions of this `Array2D`.

        Returns:
            The integer number of rows of this `Array2D`.
        """
        return len(self.__buf)

    def columns(self):
        """Gets the number of columns in this `Array2D`.

        A user should assume this runs in `O(1)` time with respect to the dimensions of this `Array2D`.

        Returns:
            The integer number of columns of this `Array2D`.
        """
        return len(self.__buf[0])

    def __iter__(self):
        # Disables the users' ability to do for e in a for an Array2D a. Without disabling this, the default Python
        # __iter__ will try to call __getitem__ with an int leading to a TypeError which is somewhat confusing.
        raise RuntimeError('Not allowed to iterate over Array2D like an iterable. Must go through `[]` with indices.')

    def __getitem__(self, index):
        row, column = self.__get_row_and_column(index)
        return self.__buf[row][column]

    def __setitem__(self, index, value):
        row, column = self.__get_row_and_column(index)
        self.__buf[row][column] = value

    def __get_row_and_column(self, index):
        row, column = index
        if row < 0 or row >= len(self.__buf):
            raise IndexError(f'row index ({row}) can\'t be < 0 or >= num_rows ({len(self.__buf)})')
        if column < 0 or column >= len(self.__buf[0]):
            raise IndexError(f'column index ({column}) can\'t be < 0 or >= num_columns ({len(self.__buf[0])})')
        return row, column


def sort(a):
    """Gets a sorted copy of an `Array`.

    This does not sort the `Array` in place. A user should assume this runs in `O(n * log(n))` time where `n` is the
    length of the input `Array`.

    Args:
        a: An `Array` to be sorted.

    Returns:
        A sorted (`Array`) copy of `a`.

    Raises:
        TypeError: If `a` is not an `Array`.
    """
    if not isinstance(a, Array):
        raise TypeError(f'can only sort Array objects')
    copy = Array(a.length())
    for i in range(a.length()):
        copy[i] = a[i]
    copy.sort()
    return copy



"""This module holds classes related to FIFO queues.

This module contains `Queue`, an implementation of a FIFO queue, and `QueueUnderflowError`, an error raised by `Queue`.

Examples:
    Initializing, adding, and removing elements from a `Queue`:

        q = Queue()
        q.enqueue(1)
        q.enqueue(2)
        q.dequeue()

    The following code will raise a `QueueUnderflowError` because the second dequeue is done on an empty `Queue`:

        q.dequeue()
        q.dequeue()
"""


class QueueUnderflowError(Exception):
    """This class is used by `Queue` to raise errors for operations done on an empty `Queue`."""

    def __init__(self, operation):
        """Initializes a `QueueUnderflowError` that will be raised associated with a particular `Queue` operation.

        Args:
            operation: a string specifying the operation to raise an error on
        """
        super().__init__(f'Cannot perform {operation} on an empty queue.')


class Queue:
    """This class represents a FIFO queue.

    One may assume that this `Queue` has no maximum capacity.

    Examples:
        To initialize a `Queue`:

            q = Queue()

        To add elements to the end of `q`:

            q.enqueue(1)
            q.enqueue(2)

        To remove and return the element at the front of `q` (in this case `x = 1`):

            x = q.dequeue()

        To see the element at the front of `q` (in this case `y = 2`):

            y = q.front()
    """

    def __init__(self):
        """Initializes an empty `Queue` in `O(1)` time."""
        self.__buf = list()

    def front(self):
        """Gets the element at the front of this `Queue`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `Queue`.

        Returns:
            The element at the front of the `Queue`. That is, the element that was first added to this `Queue` of the
            elements in it.

        Raises:
             QueueUnderflowError: If this `Queue` is empty.
        """
        if len(self.__buf) == 0:
            raise QueueUnderflowError('front()')
        return self.__buf[0]

    def dequeue(self):
        """Removes the element at the front of this `Queue`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `Queue`.

        Returns:
            The element at the front of the `Queue` that was removed. That is, the element that was first added to this
            `Queue` of the elements in it.

        Raises:
             QueueUnderflowError: If this `Queue` is empty.
        """
        if len(self.__buf) == 0:
            raise QueueUnderflowError('dequeue()')
        return self.__buf.pop(0)

    def enqueue(self, value):
        """Adds an element to the end of this `Queue`.

        One may assume that this operation runs in `O(1)` time with respect to the size of this `Queue`.

        Args:
            value: Element to add to this `Queue`. It can be of any type.
        """
        self.__buf.append(value)

    def is_empty(self):
        """Returns `True` if this `Queue` is empty, `False` otherwise in `O(1)` time w/r/t the size of this `Queue`."""
        return len(self.__buf) == 0

    def size(self):
        """Returns the integer number of elements in this `Queue` in `O(1)` time w/r/t the size of this `Queue`."""
        return len(self.__buf)



"""Module that holds classes related to hashing.

This module contains the `HashTable` class. `HashTable` represents a hash table.

Examples:
    Creating a `HashTable`, adding key-value pairs, and retrieving values:

        t = HashTable()
        t.insert('a', 1)
        t.insert('b', 2)
        t.insert('c', 3)
        x = t.get_value('a')

    Removing elements from a `HashTable`:

        t.delete('a')
"""


class HashTable:
    """Represents a hash table that stores key-value pairs.

    Examples:
        To initialize a `HashTable`:

            t = HashTable()

        To insert key-value pairs:

            t.insert('a', 1)
            t.insert('b', 2)

        To check if `t` contains a key:

            if t.contains_key('a'):
                # Do something

        To remove a key-value pair:

            t.delete('a')
    """

    def __init__(self):
        """Initializes an empty `HashTable` in `O(1)` time."""
        self.__table = dict()

    def insert(self, key, value):
        """Inserts a key-value pair into this `HashTable`.

        One should assume that this runs in `O(1)` time w/r/t the number of entries in this `HashTable`.

        Args:
            key: A key. If its a new key, then a new entry will be added to this `HashTable` a new pair is added. This
                 can be of any type.
            value: A value to be associated with `key`. If `key` is already in this `HashTable`, then it will now have
                   `value` associated with it and the number of entries in the `HashTable` will not change. This can
                   be of any type including `None`.
        """
        self.__table[key] = value

    def contains_key(self, key):
        """Returns whether this `HashTable` contains a particular key.

        One should assume that this runs in `O(1)` time w/r/t the number of entries in this `HashTable`.

        Args:
            key: The key being searched.

        Returns:
            `True` if `key` is in this `HashTable` and `False` otherwise.
        """
        return key in self.__table

    def get_value(self, key):
        """Gets the value associated with a key in the `HashTable`.

        One should assume that this runs in `O(1)` time w/r/t the number of entries in this `HashTable`.

        Args:
            key: The key whose value is being retrieved. This can be of any type.

        Returns:
            The value associated with `key` if `key` is in this `HashTable`.

        Raises:
            KeyError: If `key` is not in this `HashTable`.
        """
        if key not in self.__table:
            raise KeyError(f'{key} not in table')
        return self.__table[key]

    def delete(self, key):
        """Removes the key-value pair with a particular key in the `HashTable`.

        One should assume that this runs in `O(1)` time w/r/t the number of entries in this `HashTable`.

        Args:
            key: The key specifying which key-value pair in the table should be removed. This can be of any type.

        Returns:
            The value that was previously associated with `key` if `key` is in this `HashTable` (before the key-value
            pair was removed).

        Raises:
            KeyError: If `key` is not in this `HashTable`.
        """
        if key not in self.__table:
            raise KeyError(f'cannot delete {key} as it does not exist in table')
        return self.__table.pop(key)
